from neuralprophet.components.future_regressors.linear import LinearFutureRegressors
from neuralprophet.components.future_regressors.neural_nets import NeuralNetsFutureRegressors
from neuralprophet.components.future_regressors.shared_neural_nets import SharedNeuralNetsFutureRegressors
from neuralprophet.components.future_regressors.shared_neural_nets_coef import SharedNeuralNetsCoefFutureRegressors
from neuralprophet.components.seasonality.fourier import (
    GlobalFourierSeasonality,
    GlocalFourierSeasonality,
    LocalFourierSeasonality,
)
from neuralprophet.components.trend.linear import GlobalLinearTrend, LocalLinearTrend
from neuralprophet.components.trend.piecewise_linear import GlobalPiecewiseLinearTrend, LocalPiecewiseLinearTrend
from neuralprophet.components.trend.static import StaticTrend


def get_trend(config, n_forecasts, quantiles, id_list, num_trends_modelled, device):
    """
    Router for all trend classes.

    Based on the conditions provided, the correct trend class is returned and initialized using the provided args.

    Parameters
        ----------
            config : configure.Trend
            n_forecasts : int
                number of steps to forecast. Aka number of model outputs
            quantiles : list
                the set of quantiles estimated
            id_list : list
                List of different time series IDs, used for global-local modelling (if enabled)

                Note
                ----
                This parameter is set to  ``['__df__']`` if only one time series is input.
            num_trends_modelled : int
                Number of different trends modelled.

                Note
                ----
                If only 1 time series is modelled, it will be always 1.

                Note
                ----
                For multiple time series. If trend is modelled globally the value is set
                to 1, otherwise it is set to the number of time series modelled.
            device : torch.device
                Device that tensors are stored on.

                Note
                ----
                This is set to ``torch.device("cpu")`` if no GPU is available.
    """
    args = {
        "config": config,
        "id_list": id_list,
        "quantiles": quantiles,
        "num_trends_modelled": num_trends_modelled,
        "n_forecasts": n_forecasts,
        "device": device,
    }

    if config.growth == "off":
        # No trend
        return StaticTrend(**args)
    elif config.growth in ["linear", "discontinuous"]:
        # Linear trend
        if num_trends_modelled == 1:
            # Global trend
            if int(config.n_changepoints) == 0:
                # Linear trend
                return GlobalLinearTrend(**args)
            else:
                # Piecewise trend
                return GlobalPiecewiseLinearTrend(**args)
        else:
            # Local trend
            if int(config.n_changepoints) == 0:
                # Linear trend
                return LocalLinearTrend(**args)
            else:
                # Piecewise trend
                return LocalPiecewiseLinearTrend(**args)
    else:
        raise ValueError(f"Growth type {config.growth} is not supported.")


def get_future_regressors(config, id_list, quantiles, n_forecasts, device, config_trend_none_bool):
    """
    Router for all future regressor classes.
    """
    args = {
        "config": config,
        "id_list": id_list,
        "quantiles": quantiles,
        "n_forecasts": n_forecasts,
        "device": device,
        "config_trend_none_bool": config_trend_none_bool,
    }
    if config.model == "linear":
        return LinearFutureRegressors(**args)
    elif config.model == "neural_nets":
        return NeuralNetsFutureRegressors(**args)
    elif config.model == "shared_neural_nets":
        return SharedNeuralNetsFutureRegressors(**args)
    elif config.model == "shared_neural_nets_coef":
        return SharedNeuralNetsCoefFutureRegressors(**args)
    else:
        raise ValueError(f"Model type {config.model} is not supported.")


def get_seasonality(
    config, id_list, quantiles, num_seasonalities_modelled, num_seasonalities_modelled_dict, n_forecasts, device
):
    """
    Router for all seasonality classes.
    """
    args = {
        "config": config,
        "id_list": id_list,
        "quantiles": quantiles,
        "n_forecasts": n_forecasts,
        "device": device,
        "num_seasonalities_modelled": num_seasonalities_modelled,
        "num_seasonalities_modelled_dict": num_seasonalities_modelled_dict,
    }

    # if only 1 time series, global strategy
    if len(id_list) == 1:
        config.global_local = "global"
        for season_i in config.periods:
            config.periods[season_i].global_local = "global"

    # if all config.periods[season_i] are the same(x), then config.global_local = x
    if len(set([config.periods[season_i].global_local for season_i in config.periods])) == 1:
        config.global_local = list(set([config.periods[season_i].global_local for season_i in config.periods]))[0]

    # if all config.periods[season_i] are different, then config.global_local = 'glocal'
    if len(set([config.periods[season_i].global_local for season_i in config.periods])) > 1:
        config.global_local = "glocal"

    if config.global_local == "global":
        # Global seasonality
        return GlobalFourierSeasonality(**args)
    elif config.global_local == "local":
        # Local seasonality
        return LocalFourierSeasonality(**args)
    elif config.global_local == "glocal":
        # Glocal seasonality
        return GlocalFourierSeasonality(**args)
    else:
        raise ValueError(f"Seasonality mode {config.global_local} is not supported.")



import logging

import numpy as np

from neuralprophet.forecaster import NeuralProphet

log = logging.getLogger("NP.forecaster")


class TorchProphet(NeuralProphet):
    """
    Prophet wrapper for the NeuralProphet forecaster.

    Parameters
    ----------
    growth: String 'linear' or 'flat' to specify a linear or
        flat trend. Note: 'flat' is equivalent to 'off' in NeuralProphet.
    changepoints: List of dates at which to include potential changepoints. If
        not specified, potential changepoints are selected automatically.
    n_changepoints: Number of potential changepoints to include. Not used
        if input `changepoints` is supplied. If `changepoints` is not supplied,
        then n_changepoints potential changepoints are selected uniformly from
        the first `changepoint_range` proportion of the history.
    changepoint_range: Proportion of history in which trend changepoints will
        be estimated. Defaults to 0.8 for the first 80%. Not used if
        `changepoints` is specified.
    yearly_seasonality: Fit yearly seasonality.
        Can be 'auto', True, False, or a number of Fourier terms to generate.
    weekly_seasonality: Fit weekly seasonality.
        Can be 'auto', True, False, or a number of Fourier terms to generate.
    daily_seasonality: Fit daily seasonality.
        Can be 'auto', True, False, or a number of Fourier terms to generate.
    holidays: pd.DataFrame with columns holiday (string) and ds (date type)
        and optionally columns lower_window and upper_window which specify a
        range of days around the date to be included as holidays.
        lower_window=-2 will include 2 days prior to the date as holidays. Also
        optionally can have a column prior_scale specifying the prior scale for
        that holiday.
    seasonality_mode: 'additive' (default) or 'multiplicative'.
    seasonality_prior_scale: Not supported for regularisation in NeuralProphet,
        please use the `seasonality_reg` arg instead.
    holidays_prior_scale: Not supported for regularisation in NeuralProphet.
    changepoint_prior_scale: Not supported for regularisation in NeuralProphet,
        please use the `trend_reg` arg instead.
    mcmc_samples: Not required for NeuralProphet
    interval_width: Float, width of the uncertainty intervals provided
        for the forecast. Converted to list of quantiles for NeuralProphet. Use
        the quantiles arg to pass quantiles directly to NeuralProphet.
    uncertainty_samples: Not required for NeuralProphet.
    stan_backend: Not supported by NeuralProphet.
    """

    def __init__(
        self,
        growth="linear",
        changepoints=None,
        n_changepoints=25,
        changepoint_range=0.8,
        yearly_seasonality="auto",
        weekly_seasonality="auto",
        daily_seasonality="auto",
        holidays=None,
        seasonality_mode="additive",
        seasonality_prior_scale=None,
        holidays_prior_scale=None,
        changepoint_prior_scale=None,
        mcmc_samples=None,
        interval_width=0.80,
        uncertainty_samples=None,
        stan_backend=None,
        **kwargs,
    ):
        # Check for unsupported features
        if seasonality_prior_scale or holidays_prior_scale or changepoint_prior_scale:
            log.error(
                "Using `_prior_scale` is unsupported for regularisation in NeuralProphet, please use the corresponding "
                "`_reg` arg instead."
            )
        if mcmc_samples or uncertainty_samples:
            log.warning(
                "Providing the number of samples for Bayesian inference or Uncertainty estimation is not required in "
                "NeuralProphet."
            )
        if stan_backend:
            log.warning("A stan_backend is not used in NeuralProphet. Please remove the parameter")

        # Handle growth
        if growth == "flat":
            log.warning("Using 'flat' growth is equivalent to 'off' in NeuralProphet.")
            growth = "off"

        # Handle quantiles
        if "quantiles" not in kwargs:
            alpha = 1 - interval_width
            quantiles = [np.round(alpha / 2, 4), np.round(1 - (alpha / 2), 4)]

        # Run the NeuralProphet function
        super(TorchProphet, self).__init__(
            growth=growth,
            changepoints=changepoints,
            n_changepoints=n_changepoints,
            changepoints_range=changepoint_range,
            yearly_seasonality=yearly_seasonality,
            weekly_seasonality=weekly_seasonality,
            daily_seasonality=daily_seasonality,
            seasonality_mode=seasonality_mode,
            quantiles=quantiles,
            **kwargs,
        )
        # Handle holidays as events
        if holidays is not None:
            self.add_events(
                events=list(holidays["holiday"].unique()),
                lower_window=holidays["lower_window"].max(),
                upper_window=holidays["upper_window"].max(),
            )
            self.events_df = holidays.copy()
            self.events_df.rename(columns={"holiday": "event"}, inplace=True)
            self.events_df.drop(["lower_window", "upper_window"], axis=1, errors="ignore", inplace=True)

        # Overwrite NeuralProphet properties
        self.name = "TorchProphet"
        self.history = None

        # Unused properties
        self.train_holiday_names = None

    def validate_inputs(self):
        """
        Validates the inputs to NeuralProphet.
        """
        log.error("Not required in NeuralProphet as all inputs are automatically checked.")

    def validate_column_name(self, name, check_holidays=True, check_seasonalities=True, check_regressors=True):
        """Validates the name of a seasonality, holiday, or regressor.

        Parameters
        ----------
        name: string
        check_holidays: bool check if name already used for holiday
        check_seasonalities: bool check if name already used for seasonality
        check_regressors: bool check if name already used for regressor
        """
        super(TorchProphet, self)._validate_column_name(
            name=name,
            events=check_holidays,
            seasons=check_seasonalities,
            regressors=check_regressors,
            covariates=check_regressors,
        )

    def setup_dataframe(self, df, initialize_scales=False):
        """
        Dummy function that raises an error.

        This function is not supported in NeuralProphet.
        """
        log.error(
            "Not required in NeuralProphet as the dataframe is automatically prepared using the private `_normalize` "
            "function."
        )

    def fit(self, df, **kwargs):
        """Fit the NeuralProphet model.

        This sets self.params to contain the fitted model parameters. It is a
        dictionary parameter names as keys and the following items:
            k (Mx1 array): M posterior samples of the initial slope.
            m (Mx1 array): The initial intercept.
            delta (MxN array): The slope change at each of N changepoints.
            beta (MxK matrix): Coefficients for K seasonality features.
            sigma_obs (Mx1 array): Noise level.
        Note that M=1 if MAP estimation.

        Parameters
        ----------
        df: pd.DataFrame containing the history. Must have columns ds (date
            type) and y, the time series. If self.growth is 'logistic', then
            df must also have a column cap that specifies the capacity at
            each ds.
        kwargs: Additional arguments passed to the optimizing or sampling
            functions in Stan.

        Returns
        -------
        The fitted NeuralProphet object.
        """
        # Check for unsupported features
        if "cap" in df.columns:
            raise NotImplementedError("Saturating forecasts using cap is not supported in NeuralProphet.")
        if "show_progress" in kwargs:
            del kwargs["show_progress"]
        # Handle holidays as events
        if hasattr(self, "events_df"):
            df = self.create_df_with_events(df, self.events_df)
        # Run the NeuralProphet function
        metrics_df = super(TorchProphet, self).fit(df=df, **kwargs)
        # Store the df for future use like in Prophet
        self.history = df
        return metrics_df

    def predict(self, df=None, **kwargs):
        """Predict using the NeuralProphet model.

        Parameters
        ----------
        df: pd.DataFrame with dates for predictions (column ds), and capacity
            (column cap) if logistic growth. If not provided, predictions are
            made on the history.

        Returns
        -------
        A pd.DataFrame with the forecast components.
        """
        if df is None:
            df = self.history.copy()
        df = super(TorchProphet, self).predict(df=df, **kwargs)
        for column in df.columns:
            # Copy column according to Prophet naming convention
            if "event_" in column:
                df[column.replace("event_", "")] = df[column]
        return df

    def predict_trend(self, df):
        """Predict trend using the NeuralProphet model.

        Parameters
        ----------
        df: Prediction dataframe.

        Returns
        -------
        Vector with trend on prediction dates.
        """
        df = super(TorchProphet, self).predict_trend(self, df, quantile=0.5)
        return df["trend"].to_numpy()

    def make_future_dataframe(self, periods, freq="D", include_history=True, **kwargs):
        """Simulate the trend using the extrapolated generative model.

        Parameters
        ----------
        periods: Int number of periods to forecast forward.
        freq: Any valid frequency for pd.date_range, such as 'D' or 'M'.
        include_history: Boolean to include the historical dates in the data
            frame for predictions.

        Returns
        -------
        pd.Dataframe that extends forward from the end of self.history for the
        requested number of periods.
        """
        # Convert all frequencies to daily
        if freq == "M":
            periods = periods * 30
        # Run the NeuralProphet function
        if hasattr(self, "events_df"):
            # Pass holidays as events
            df_future = super(TorchProphet, self).make_future_dataframe(
                df=self.history,
                events_df=self.events_df,
                periods=periods,
                n_historic_predictions=include_history,
                **kwargs,
            )
        else:
            df_future = super(TorchProphet, self).make_future_dataframe(
                df=self.history, periods=periods, n_historic_predictions=include_history, **kwargs
            )
        return df_future

    def add_seasonality(self, name, period, fourier_order, prior_scale=None, mode=None, condition_name=None, **kwargs):
        """Add a seasonal component with specified period, number of Fourier
        components, and prior scale.

        Increasing the number of Fourier components allows the seasonality to
        change more quickly (at risk of overfitting). Default values for yearly
        and weekly seasonalities are 10 and 3 respectively.

        Increasing prior scale will allow this seasonality component more
        flexibility, decreasing will dampen it. If not provided, will use the
        seasonality_prior_scale provided on initialization (defaults
        to 10).

        Mode can be specified as either 'additive' or 'multiplicative'. If not
        specified, self.seasonality_mode will be used (defaults to additive).
        Additive means the seasonality will be added to the trend,
        multiplicative means it will multiply the trend.

        If condition_name is provided, the dataframe passed to `fit` and
        `predict` should have a column with the specified condition_name
        containing booleans which decides when to apply seasonality.

        Parameters
        ----------
        name: string name of the seasonality component.
        period: float number of days in one period.
        fourier_order: int number of Fourier components to use.
        prior_scale: Not supported in NeuralProphet.
        mode: optional 'additive' or 'multiplicative'
        condition_name: Not supported in NeuralProphet.

        Returns
        -------
        The NeuralProphet object.
        """
        # Check for unsupported features
        if condition_name:
            raise NotImplementedError("Conditioning on seasonality is not supported in NeuralProphet.")
        if prior_scale:
            log.warning(
                "Prior scale is not supported in NeuralProphet. Use the `regularisation` parameter for regularisation."
            )
        # Set attributes in NeuralProphet config
        try:
            self.season_config.mode = mode
        except AttributeError:
            log.warning("Cannot set the seasonality mode attribute in NeuralProphet. Pleas inspect manually.")
        # Run the NeuralProphet function
        return super(TorchProphet, self).add_seasonality(name, period, fourier_order, **kwargs)

    def add_regressor(self, name, prior_scale=None, standardize="auto", mode="additive", **kwargs):
        """Add an additional (future) regressor to be used for fitting and predicting.

        Parameters
        ----------
        name: string name of the regressor.
        prior_scale: Not supported in NeuralProphet.
        standardize: optional, specify whether this regressor will be
            standardized prior to fitting. Can be 'auto' (standardize if not
            binary), True, or False.
        mode: optional, 'additive' or 'multiplicative'. Defaults to
            self.seasonality_mode. Not supported in NeuralProphet.

        Returns
        -------
        The NeuralProphet object.
        """
        # Check for unsupported features
        if prior_scale:
            log.warning(
                "Prior scale is not supported in NeuralProphet. Use the `regularisation` parameter for regularisation."
            )
        # Run the NeuralProphet function
        super(TorchProphet, self).add_future_regressor(name, normalize=standardize, **kwargs)
        return self

    def add_country_holidays(self, country_name, **kwargs):
        """Add in built-in holidays for the specified country.

        These holidays will be included in addition to any specified on model
        initialization.

        Holidays will be calculated for arbitrary date ranges in the history
        and future. See the online documentation for the list of countries with
        built-in holidays.

        Built-in country holidays can only be set for a single country.

        Parameters
        ----------
        country_name: Name of the country, like 'UnitedStates' or 'US'

        Returns
        -------
        The NeuralProphet object.
        """
        super(TorchProphet, self).add_country_holidays(country_name=country_name, **kwargs)

    def plot(
        self,
        fcst,
        ax=None,
        uncertainty=True,
        plot_cap=True,
        xlabel="ds",
        ylabel="y",
        figsize=(10, 6),
        include_legend=False,
        **kwargs,
    ):
        """Plot the NeuralProphet forecast.

        Parameters
        ----------
        fcst: pd.DataFrame output of self.predict.
        ax: Optional matplotlib axes on which to plot.
        uncertainty: Not supported in NeuralProphet.
        plot_cap: Not supported in NeuralProphet.
        xlabel: Optional label name on X-axis
        ylabel: Optional label name on Y-axis
        figsize: Optional tuple width, height in inches.
        include_legend: Not supported in NeuralProphet.

        Returns
        -------
        A matplotlib figure.
        """
        log.warning("The attributes `uncertainty`, `plot_cap` and `include_legend` are not supported by NeuralProphet")
        fig = super(TorchProphet, self).plot(fcst=fcst, ax=ax, xlabel=xlabel, ylabel=ylabel, figsize=figsize, **kwargs)
        return fig

    def plot_components(
        self, fcst, uncertainty=True, plot_cap=True, weekly_start=0, yearly_start=0, figsize=None, **kwargs
    ):
        """Plot the NeuralProphet forecast components.

        Will plot whichever are available of: trend, holidays, weekly
        seasonality, and yearly seasonality.

        Parameters
        ----------
        fcst: pd.DataFrame output of self.predict.
        uncertainty: Not supported in NeuralProphet.
        plot_cap: Not supported in NeuralProphet.
        weekly_start: Not supported in NeuralProphet.
        yearly_start: Not supported in NeuralProphet.
        figsize: Optional tuple width, height in inches.

        Returns
        -------
        A matplotlib figure.
        """
        log.warning(
            "The attributes `uncertainty`, `plot_cap`, `weekly_start` and `yearly_start` are not supported by "
            "NeuralProphet"
        )
        fig = super(TorchProphet, self).plot_components(fcst=fcst, figsize=figsize, **kwargs)
        return fig


def plot(
    self,
    fcst,
    ax=None,
    uncertainty=True,
    plot_cap=True,
    xlabel="ds",
    ylabel="y",
    figsize=(10, 6),
    include_legend=False,
    **kwargs,
):
    """Plot the NeuralProphet forecast.

    Parameters
    ----------
    fcst: pd.DataFrame output of self.predict.
    ax: Optional matplotlib axes on which to plot.
    uncertainty: Not supported in NeuralProphet.
    plot_cap: Not supported in NeuralProphet.
    xlabel: Optional label name on X-axis
    ylabel: Optional label name on Y-axis
    figsize: Optional tuple width, height in inches.
    include_legend: Not supported in NeuralProphet.

    Returns
    -------
    A matplotlib figure.
    """
    log.warning("The attributes `uncertainty`, `plot_cap` and `include_legend` are not supported by NeuralProphet")
    fig = super(TorchProphet, self).plot(fcst=fcst, ax=ax, xlabel=xlabel, ylabel=ylabel, figsize=figsize, **kwargs)
    return fig


def plot_plotly(
    self,
    fcst,
    ax=None,
    uncertainty=True,
    plot_cap=True,
    xlabel="ds",
    ylabel="y",
    figsize=(10, 6),
    include_legend=False,
    **kwargs,
):
    """Plot the NeuralProphet forecast.

    Parameters
    ----------
    fcst: pd.DataFrame output of self.predict.
    ax: Optional matplotlib axes on which to plot.
    uncertainty: Not supported in NeuralProphet.
    plot_cap: Not supported in NeuralProphet.
    xlabel: Optional label name on X-axis
    ylabel: Optional label name on Y-axis
    figsize: Optional tuple width, height in inches.
    include_legend: Not supported in NeuralProphet.

    Returns
    -------
    A matplotlib figure.
    """
    log.warning("The attributes `uncertainty`, `plot_cap` and `include_legend` are not supported by NeuralProphet")
    fig = super(TorchProphet, self).plot(
        fcst=fcst, ax=ax, xlabel=xlabel, ylabel=ylabel, figsize=figsize, plotting_backend="plotly", **kwargs
    )
    return fig


def plot_components(m, fcst, uncertainty=True, plot_cap=True, weekly_start=0, yearly_start=0, figsize=None, **kwargs):
    """
    Plot the NeuralProphet forecast components.

    Will plot whichever are available of: trend, holidays, weekly
    seasonality, yearly seasonality, and additive and multiplicative extra
    regressors.

    Parameters
    ----------
    m: NeuralProphet model.
    fcst: pd.DataFrame output of m.predict.
    uncertainty: Not supported in NeuralProphet.
    plot_cap: Not supported in NeuralProphet.
    weekly_start: Not supported in NeuralProphet.
    yearly_start: Not supported in NeuralProphet.
    figsize: Optional tuple width, height in inches.

    Returns
    -------
    A matplotlib figure.
    """
    log.warning(
        "The attributes `uncertainty`, `plot_cap`, `weekly_start` and `yearly_start` are not supported by NeuralProphet"
    )
    # Run the NeuralProphet plotting function
    fig = m.plot_components(fcst, **kwargs)
    return fig


def plot_components_plotly(m, fcst, uncertainty=True, plot_cap=True, figsize=(900, 200), **kwargs):
    """
    Plot the NeuralProphet forecast components using Plotly.
    See plot_plotly() for Plotly setup instructions

    Will plot whichever are available of: trend, holidays, weekly
    seasonality, yearly seasonality, and additive and multiplicative extra
    regressors.

    Parameters
    ----------
    m: NeuralProphet model.
    fcst: pd.DataFrame output of m.predict.
    uncertainty: Not supported in NeuralProphet.
    plot_cap: Not supported in NeuralProphet.
    figsize: Not supported in NeuralProphet.
    Returns
    -------
    A Plotly Figure.
    """
    log.warning(
        "The attributes `uncertainty`, `plot_cap`, `weekly_start` and `yearly_start` are not supported by NeuralProphet"
    )
    # Run the NeuralProphet plotting function
    fig = m.plot_components(fcst, figsize=None, plotting_backend="plotly", **kwargs)
    return fig



import re
from dataclasses import dataclass
from typing import Any, List, Tuple, Union

import matplotlib
import numpy as np
import pandas as pd

from neuralprophet.plot_forecast_matplotlib import plot_interval_width_per_timestep, plot_nonconformity_scores
from neuralprophet.plot_forecast_plotly import (
    plot_interval_width_per_timestep as plot_interval_width_per_timestep_plotly,
)
from neuralprophet.plot_forecast_plotly import plot_nonconformity_scores as plot_nonconformity_scores_plotly
from neuralprophet.plot_utils import log_warning_deprecation_plotly, select_plotting_backend


@dataclass
class Conformal:
    """Conformal prediction dataclass

    Parameters
    ----------
    alpha : float or tuple
        user-specified significance level of the prediction interval, float if coverage error spread arbitrarily over
        left and right tails, tuple of two floats for different coverage error over left and right tails respectively
    method : str
        name of conformal prediction technique used

        Options
            * ``naive``: Naive or Absolute Residual
            * ``cqr``: Conformalized Quantile Regression
    n_forecasts : int
        optional, number of steps ahead of prediction time step to forecast
    quantiles : list
        optional, list of quantiles for quantile regression uncertainty estimate

    """

    alpha: Union[float, Tuple[float, float]]
    method: str
    n_forecasts: int
    quantiles: List[float]

    def __post_init__(self):
        if isinstance(self.alpha, float):
            self.symmetrical = True
            self.q_hats = pd.DataFrame(columns=["q_hat_sym"])
        elif self.method == "naive":
            raise ValueError(
                "Asymmetrical coverage errors are not available for the naive method. "
                "Please use one alpha or method='cqr'."
            )
        else:
            self.symmetrical = False
            self.alpha_lo, self.alpha_hi = self.alpha
            self.q_hats = pd.DataFrame(columns=["q_hat_lo", "q_hat_hi"])
        self.noncon_scores = dict()

    def predict(self, df: pd.DataFrame, df_cal: pd.DataFrame, show_all_PI: bool = False) -> pd.DataFrame:
        """Apply a given conformal prediction technique to get the uncertainty prediction intervals (or q-hat) for test
        dataframe.

        Parameters
        ----------
            df : pd.DataFrame
                test dataframe
            df_cal : pd.DataFrame
                calibration dataframe
            show_all_PI : bool
                whether to return all prediction intervals (including quantile regression and conformal prediction)

            Returns
            -------
                pd.DataFrame
                    test dataframe with uncertainty prediction intervals

        """
        df_qr = df.copy()

        for step_number in range(1, self.n_forecasts + 1):
            # conformalize
            noncon_scores = self._get_nonconformity_scores(df_cal, step_number)
            q_hat = self._get_q_hat(df_cal, noncon_scores)
            y_hat_col = f"yhat{step_number}"
            y_hat_lo_col = f"{y_hat_col} {min(self.quantiles) * 100}%"
            y_hat_hi_col = f"{y_hat_col} {max(self.quantiles) * 100}%"
            if self.method == "naive" and self.symmetrical:
                q_hat_sym = q_hat["q_hat_sym"]
                df[y_hat_lo_col] = df[y_hat_col] - q_hat_sym
                df[y_hat_hi_col] = df[y_hat_col] + q_hat_sym
            elif self.method == "cqr" and self.symmetrical:
                q_hat_sym = q_hat["q_hat_sym"]
                df[y_hat_lo_col] = df[y_hat_lo_col] - q_hat_sym
                df[y_hat_hi_col] = df[y_hat_hi_col] + q_hat_sym
            elif self.method == "cqr" and not self.symmetrical:
                q_hat_lo = q_hat["q_hat_lo"]
                q_hat_hi = q_hat["q_hat_hi"]
                df[y_hat_lo_col] = df[y_hat_lo_col] - q_hat_lo
                df[y_hat_hi_col] = df[y_hat_hi_col] + q_hat_hi
            else:
                raise ValueError(
                    f"Unknown conformal prediction method '{self.method}'. Please input either 'naive' or 'cqr'."
                )
            if step_number == 1:
                # save nonconformity scores of the first timestep
                self.noncon_scores = noncon_scores

            # append the dictionary of q_hats to the dataframe based on the keys of the dictionary
            q_hat_df = pd.DataFrame([q_hat])
            self.q_hats = pd.concat([self.q_hats, q_hat_df], ignore_index=True)

            # if show_all_PI is True, add the quantile regression prediction intervals
            if show_all_PI:
                df_quantiles = [col for col in df_qr.columns if "%" in col and f"yhat{step_number}" in col]
                df_add = df_qr[df_quantiles]

                if self.method == "naive":
                    cp_lo_col = f"yhat{step_number} - qhat{step_number}"  # e.g. yhat1 - qhat1
                    cp_hi_col = f"yhat{step_number} + qhat{step_number}"  # e.g. yhat1 + qhat1
                    df.rename(columns={y_hat_lo_col: cp_lo_col, y_hat_hi_col: cp_hi_col}, inplace=True)
                elif self.method == "cqr":
                    qr_lo_col = (
                        f"yhat{step_number} {max(self.quantiles) * 100}% - qhat{step_number}"  # e.g. yhat1 95% - qhat1
                    )
                    qr_hi_col = (
                        f"yhat{step_number} {min(self.quantiles) * 100}% + qhat{step_number}"  # e.g. yhat1 5% + qhat1
                    )
                    df.rename(columns={y_hat_lo_col: qr_lo_col, y_hat_hi_col: qr_hi_col}, inplace=True)

                df = pd.concat([df, df_add], axis=1, ignore_index=False)

        return df

    def _get_nonconformity_scores(self, df_cal: pd.DataFrame, step_number: int) -> dict:
        """Get the nonconformity scores using the given conformal prediction technique.

        Parameters
        ----------
            df_cal : pd.DataFrame
                calibration dataframe
            step_number : int
                i-th step ahead forecast

            Returns
            -------
                Dict[str, np.ndarray]
                    dictionary with one entry (symmetrical) or two entries (asymmetrical) of nonconformity scores

        """
        y_hat_col = f"yhat{step_number}"
        if self.method == "cqr":
            # CQR nonconformity scoring function
            quantile_lo = str(min(self.quantiles) * 100)
            quantile_hi = str(max(self.quantiles) * 100)
            quantile_lo_col = f"{y_hat_col} {quantile_lo}%"
            quantile_hi_col = f"{y_hat_col} {quantile_hi}%"
            if self.symmetrical:

                def cqr_scoring_func(row):
                    return (
                        [None, None]
                        if row[quantile_lo_col] is None or row[quantile_hi_col] is None
                        else [
                            max(row[quantile_lo_col] - row["y"], row["y"] - row[quantile_hi_col]),
                            0 if row[quantile_lo_col] - row["y"] > row["y"] - row[quantile_hi_col] else 1,
                        ]
                    )

                scores_df = df_cal.apply(cqr_scoring_func, axis=1, result_type="expand")
                scores_df.columns = ["scores", "arg"]
                noncon_scores = scores_df["scores"].values
            else:  # asymmetrical intervals

                def cqr_scoring_func(row):
                    return (
                        [None, None]
                        if row[quantile_lo_col] is None or row[quantile_hi_col] is None
                        else [
                            row[quantile_lo_col] - row["y"],
                            row["y"] - row[quantile_hi_col],
                            0 if row[quantile_lo_col] - row["y"] > row["y"] - row[quantile_hi_col] else 1,
                        ]
                    )

                scores_df = df_cal.apply(cqr_scoring_func, axis=1, result_type="expand")
                scores_df.columns = ["scores_lo", "scores_hi", "arg"]
                noncon_scores_lo = scores_df["scores_lo"].values
                noncon_scores_hi = scores_df["scores_hi"].values
                # Remove NaN values
                noncon_scores_lo: Any = noncon_scores_lo[~pd.isnull(noncon_scores_lo)]
                noncon_scores_hi: Any = noncon_scores_hi[~pd.isnull(noncon_scores_hi)]
                # Sort
                noncon_scores_lo.sort()
                noncon_scores_hi.sort()
                # return dict of nonconformity scores
                return {"noncon_scores_hi": noncon_scores_lo, "noncon_scores_lo": noncon_scores_hi}
        else:  # self.method == "naive"
            # Naive nonconformity scoring function
            noncon_scores = abs(df_cal["y"] - df_cal[y_hat_col]).values
        # Remove NaN values
        noncon_scores: Any = noncon_scores[~pd.isnull(noncon_scores)]
        # Sort
        noncon_scores.sort()

        return {"noncon_scores": noncon_scores}

    def _get_q_hat(self, df_cal: pd.DataFrame, noncon_scores: dict) -> dict:
        """Get the q_hat that is derived from the nonconformity scores.

        Parameters
        ----------
            df_cal : pd.DataFrame
                calibration dataframe
            noncon_scores : dict
                dictionary with one entry (symmetrical) or two entries (asymmetrical) of nonconformity scores

            Returns
            -------
                Dict[str, float]
                    upper and lower q_hat value, or the one-sided prediction interval width

        """
        # Get the q-hat index and value
        if self.method == "cqr" and self.symmetrical is False:
            noncon_scores_lo = noncon_scores["noncon_scores_lo"]
            noncon_scores_hi = noncon_scores["noncon_scores_hi"]
            q_hat_idx_lo = int(len(noncon_scores_lo) * self.alpha_lo)
            q_hat_idx_hi = int(len(noncon_scores_hi) * self.alpha_hi)
            q_hat_lo = noncon_scores_lo[-q_hat_idx_lo]
            q_hat_hi = noncon_scores_hi[-q_hat_idx_hi]
            return {"q_hat_lo": q_hat_lo, "q_hat_hi": q_hat_hi}
        else:
            noncon_scores = noncon_scores["noncon_scores"]
            q_hat_idx = int(len(noncon_scores) * self.alpha)
            q_hat = noncon_scores[-q_hat_idx]
            return {"q_hat_sym": q_hat}

    def plot(self, plotting_backend=None):
        """Apply a given conformal prediction technique to get the uncertainty prediction intervals (or q-hats).

        Parameters
        ----------
            plotting_backend : str
                specifies the plotting backend for the nonconformity scores plot, if any

                Options
                * ``plotly-resampler``: Use the plotly backend for plotting in resample mode. This mode uses the
                    plotly-resampler package to accelerate visualizing large data by resampling it. For some
                    environments (colab, pycharm interpreter) plotly-resampler might not properly vizualise the figures.
                    In this case, consider switching to 'plotly-auto'.
                * ``plotly``: Use the plotly backend for plotting
                * ``matplotlib``: use matplotlib for plotting
                * (default) None: Plotting backend ist set automatically. Use plotly with resampling for jupyterlab
                    notebooks and vscode notebooks. Automatically switch to plotly without resampling for all other
                    environments.

        """
        method = self.method.upper() if "cqr" in self.method.lower() else self.method.title()
        # Check whether a local or global plotting backend is set.
        plotting_backend = select_plotting_backend(model=self, plotting_backend=plotting_backend)

        log_warning_deprecation_plotly(plotting_backend)
        initial_q_hat = (
            self.q_hats["q_hat_sym"][0]
            if self.symmetrical
            else [self.q_hats["q_hat_lo"][0], self.q_hats["q_hat_hi"][0]]
        )
        if plotting_backend.startswith("plotly"):
            if self.n_forecasts == 1:
                fig = plot_nonconformity_scores_plotly(
                    self.noncon_scores,
                    self.alpha,
                    initial_q_hat,
                    method,
                    resampler_active=plotting_backend == "plotly-resampler",
                )
            else:
                fig = plot_interval_width_per_timestep_plotly(self.q_hats, method, resampler_active=False)
            fig.show()
        else:
            if self.n_forecasts == 1:
                # includes nonconformity scores of the first timestep
                fig = plot_nonconformity_scores(self.noncon_scores, self.alpha, initial_q_hat, method)
            else:
                fig = plot_interval_width_per_timestep(self.q_hats, method)
        if plotting_backend in ["matplotlib", "plotly", "plotly-resampler"] and matplotlib.is_interactive():
            fig


def uncertainty_evaluate(df_forecast: pd.DataFrame) -> pd.DataFrame:
    """Evaluate conformal prediction on test dataframe.

    Parameters
    ----------
        df_forecast : pd.DataFrame
            forecast dataframe with the conformal prediction intervals

    Returns
    -------
        pd.DataFrame
            table containing evaluation metrics such as interval_width and miscoverage_rate
    """
    # Remove beginning rows used as lagged regressors (if any), or future dataframes without y-values
    # therefore, this ensures that all forecast rows for evaluation contains both y and y-hat
    df_forecast_eval = df_forecast.dropna(subset=["y", "yhat1"]).reset_index(drop=True)

    # Get evaluation params
    df_eval = pd.DataFrame()
    cols = df_forecast_eval.columns
    yhat_cols = [col for col in cols if "%" in col]
    n_forecasts = int(re.search("yhat(\\d+)", yhat_cols[-1]).group(1))

    # get the highest and lowest quantile percentages
    quantiles = []
    for col in yhat_cols:
        match = re.search(r"\d+\.\d+", col)
        if match:
            quantiles.append(float(match.group()))
    quantiles = sorted(set(quantiles))

    # Begin conformal evaluation steps
    for step_number in range(1, n_forecasts + 1):
        y = df_forecast_eval["y"].values
        # only relevant if show_all_PI is true
        if len([col for col in cols if "qhat" in col]) > 0:
            qhat_cols = [col for col in cols if f"qhat{step_number}" in col]
            yhat_lo = df_forecast_eval[qhat_cols[0]].values
            yhat_hi = df_forecast_eval[qhat_cols[-1]].values
        else:
            yhat_lo = df_forecast_eval[f"yhat{step_number} {quantiles[0]}%"].values
            yhat_hi = df_forecast_eval[f"yhat{step_number} {quantiles[-1]}%"].values
        interval_width, miscoverage_rate = _get_evaluate_metrics_from_dataset(y, yhat_lo, yhat_hi)

        # Construct row dataframe with current timestep using its q-hat, interval width, and miscoverage rate
        col_names = ["interval_width", "miscoverage_rate"]
        row = [interval_width, miscoverage_rate]
        df_row = pd.DataFrame([row], columns=pd.MultiIndex.from_product([[f"yhat{step_number}"], col_names]))

        # Add row dataframe to overall evaluation dataframe with all forecasted timesteps
        df_eval = pd.concat([df_eval, df_row], axis=1)

    return df_eval


def _get_evaluate_metrics_from_dataset(y: np.ndarray, yhat_lo: np.ndarray, yhat_hi: np.ndarray) -> Tuple[float, float]:
    #     df_forecast_eval: pd.DataFrame,
    #     quantile_lo_col: str,
    #     quantile_hi_col: str,
    # ) -> Tuple[float, float]:
    """Infers evaluation parameters based on the evaluation dataframe columns.

    Parameters
    ----------
        df_forecast_eval : pd.DataFrame
            forecast dataframe with the conformal prediction intervals

    Returns
    -------
        float, float
            conformal prediction evaluation metrics
    """
    # Interval width (efficiency metric)
    quantile_lo_mean = np.mean(yhat_lo)
    quantile_hi_mean = np.mean(yhat_hi)
    interval_width = quantile_hi_mean - quantile_lo_mean

    # Miscoverage rate (validity metric)
    n_covered = np.sum((y >= yhat_lo) & (y <= yhat_hi))
    coverage_rate = n_covered / len(y)
    miscoverage_rate = 1 - coverage_rate

    return interval_width, miscoverage_rate
